# wanha - your pesky URL-guardian for IRC
# Copyright (C) 2006-2007 Tommi Saviranta  <wnd@iki.fi>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


package Wanha::Lang;

use strict;
use warnings FATAL => 'all';
use diagnostics;

use Wanha::Config;


BEGIN {
	my $lang = $Wanha::Config::lang;

	if (defined($lang)) {
		require "lang_$lang.pl";
	} else {
		my $l = $ENV{"HTTP_ACCEPT_LANGUAGE"};
		my @langs;

		if (defined($l)) {
			@langs = split(/,/, $l);
			s/;.*// foreach (@langs);
			s/-/_/ foreach (@langs);
		}
		push(@langs, "en");

		for (my $i = 0; $i <= $#langs; $i++) {
			my $file = "lang_" . $langs[$i] . ".pl";
			next unless -e "$file";
			require "$file";
			$i = $#langs;
		}
	}
}



=item get
Get string

0: id
r: string
=cut
sub get($) {
	my ($id) = @_;
	return &RLang::get($id);
}


=item printf
Return printf "format", @_

0: format with key in <&key&>
@: parameters
r: string
=cut
sub printf($@) {
	my $fmt = shift;
	while ($fmt =~ /<&.*?&>/) {
		my ($key) = $fmt =~ /<&(.*?)&>/;
		if (! defined($key)) {
			print STDERR "Wanha::Lang::printf: tag not found: $fmt\n";
			return $fmt;
		}
		my $val = &get($key);
		$fmt =~ s/<&$key&>/$val/;
	}
	printf($fmt, @_);
}



return 1;
