// ==UserScript==
// @name	Wanha links
// @namespace	http://katei.fi/wanha/
// @description	Creates a link to open wanha's "latest links" in tabs.
// @include	http://katei.fi/wanha/*
// =/UserScript==


function get_table()
{
	var divs = document.getElementsByTagName("table");
	if (divs == null) {
		GM_log("No table found -- aborting...");
		return null;
	}

	for (var i = 0; i < divs.length; i++) {
		var div = divs[i];
		if (div.getAttribute("class") == "urls") {
			return div;
		}
	}

	return null;
}


function get_base_url()
{
	var base = window.location.href;
	var sep = base.lastIndexOf("/");
	if (sep == -1) {
		// Also work with MS-stuff. No idea if this is necessary.
		sep = base.lastIndexOf("\\");
	}
	if (sep == -1) {
		return null;
	}
	
	return base.substring(0, sep + 1);
}


function get_links()
{
	var table = get_table();
	if (table == null) {
		GM_log("No URL-table found -- aborting...");
		return null;
	}

	var raw_links = table.getElementsByTagName("a");
	var links = new Array();

	var base_url = get_base_url();
	if (base_url == null) {
		GM_log("PANIC: Can't find base URL");
		return null;
	}
	base_len = base_url.length;

	var j = 0;
	for (var i = 0; i < raw_links.length; i++) {
		var link = raw_links[i];
		var url = link.getAttribute("href");
		if (url.substring(0, 4) != "http"
				|| url.substring(0, base_len) == base_url) {
			continue;
		}
		links[j++] = url;
	}

	return links;
}


function open_links()
{
	var links = get_links();

	for (var i = 0; i < links.length; i++) {
		GM_openInTab(links[i]);
	}
}


function insert_after(new_node, ref_node)
{
	var next_node = ref_node.nextSibling;
	var parent_node = ref_node.parentNode;
	if (parent_node == null) {
		GM_log("No parent node found for " + ref_node.toString());
		return;
	}
	if (next_node == null) {
		parent_node.appendChild(new_node);
	} else {
		parent_node.insertBefore(new_node, next_node);
	}
}


function print_links(after, links)
{
	for (var i = 0; i < links.length; i++) {
		var t = document.createTextNode(links[i].toString());
		insert_after(t, after);
		after = t;
		t = document.createElement("br");
		insert_after(t, after);
		after = t;
	}
}


function remove_greasemonkey_link()
{
	var p = document.getElementById("greasemonkey");
	if (p == null) {
		return;
	}

	if (p.parentNode == null) {
		return;
	}

	p.parentNode.removeChild(p);
}


// Check support for DOM (probably unnecessary)
if (! document.getElementById || ! document.createElement
		|| ! document.appendChild) {
	return;
}

// Remove Greasemonkey link
remove_greasemonkey_link();

// Check document
var table = get_table();
if (table == null) {
	GM_log("No URL table found -- aborting...");
	return;
}

// Any link here?
var links = get_links();
if (links == null || links.length == 0) {
	return;
}

// Open all
var link = document.createElement("a");
link.setAttribute("href", "#");
link.addEventListener("click", open_links, false);
// TODO: l10n
link.appendChild(document.createTextNode("Open all links in new tabs"));
var p = document.createElement("p");
p.appendChild(link);
insert_after(p, table);

// print_links(p, links);
